/*----------------------------------------------------------------------
  File    : clx.c
  Contents: probabilistic and fuzzy cluster induction
  Author  : Christian Borgelt
  History : 2003.01.30 file created
            2003.01.31 cluster index output added
            2003.03.11 hard cluster assigment added
            2003.05.17 normalization parameters moved to cluster file
            2003.08.16 slight changes in error message output
            2004.02.25 source files clx.c und mclx.c combined
            2004.04.22 bug concerning additional field reading fixed
            2006.01.28 conditional compilation simplified
            2007.01.22 alignment of membership degrees added
            2007.02.14 adapted to changed module tabread
            2007.03.16 table and matrix version combined
            2008.03.01 redesign for feature weighting finished
            2012.05.02 adapted to modified attset and utility modules
            2013.06.08 missing flag AS_MARKED added to table read mode
            2013.08.16 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.20 output format changed to significant digits
            2013.08.29 adapted to new function as_target()
            2013.08.30 missing deallocation of vector buffer added
            2014.10.24 changed from LGPL license to MIT license
            2018.08.18 extended functions (attset/table) removed
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#include "attmap.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#ifndef CLS_PARSE
#define CLS_PARSE
#endif
#include "cluster.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

#ifdef _MSC_VER
#ifndef snprintf
#define snprintf _snprintf
#endif
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "clx"
#define DESCRIPTION "probabilistic and fuzzy cluster execution"
#define VERSION     "version 4.3 (2022.04.13)         " \
                    "(c) 2003-2022   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse errors on input file */
#define E_ATTCNT    (-10)       /* no usable attributes found */
#define E_INPCNT    (-11)       /* no usable inputs found */
#define E_VECSIZE   (-12)       /* invalid data point/vector size */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*--------------------------------------------------------------------*/
#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

#if DIMID!=ATTID
#error "DIMID and ATTID must be the same data type"
#endif

#if DIMID!=TPLID
#error "DIMID and TPLID must be the same data type"
#endif

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- prediction result --- */
  DIMID  cnt;                   /* number of clusters */
  DIMID  cls;                   /* predicted cluster */
  CCHAR *col_cls;               /* name   of cluster    column */
  int    cwd_cls;               /* width  of cluster    column */
  CCHAR *col_msd;               /* name   of membership column(s) */
  int    cwd_msd;               /* width  of membership column(s) */
  int    dig_msd;               /* format of membership column(s) */
} RESULT;                       /* (prediction result) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_ATTCNT  -10 */  "no (usable) attributes (need at least 1)",
  /* E_INPCNT  -11 */  "no (usable) inputs (need at least 1)",
  /* E_VECSIZE -12 */  "invalid vector size %"DIMID_FMT,
  /*           -13 */  "unknown error",
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR    *prgname;       /* program name for error messages */
static SCANNER  *scan   = NULL; /* scanner for cluster set desc. */
static TABREAD  *tread  = NULL; /* table reader */
static TABWRITE *twrite = NULL; /* table writer */
static ATTSET   *attset = NULL; /* attribute set */
static ATTMAP   *attmap = NULL; /* attribute map */
static TABLE    *table  = NULL; /* data table */
static CLSET    *clset  = NULL; /* cluster set */
static RESULT   res     = {     /* prediction result */
  0, 0,                         /* number of clusters, pred. cluster */
  NULL, 0,                      /* data for cluster    column */
  NULL, 0, 6 };                 /* data for membership column(s) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (clset)  cls_delete(clset);     \
  if (table)  tab_delete(table,  0); \
  if (attmap) am_delete(attmap,  0); \
  if (attset) as_delete(attset);     \
  if (tread)  trd_delete(tread,  1); \
  if (twrite) twr_delete(twrite, 1); \
  if (scan)   scn_delete(scan,   1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void infout (ATTSET *set, TABWRITE *twrite, int mode)
{                               /* --- write additional information */
  DIMID c;                      /* loop variable for clusters */
  int   n, k, m;                /* character counters */
  char  b[64];                  /* buffer for number output */

  assert(set && twrite);        /* check the function arguments */
  if (res.col_cls) {            /* if to write a cluster index */
    if (mode & AS_ATT) {        /* if to write the header */
      twr_puts(twrite, res.col_cls);  /* write cluster column name */
      if ((mode & AS_ALIGN)     /* if to align the column */
      && ((mode & AS_WEIGHT) || res.col_msd)) {
        n = (int)strlen(res.col_cls);
        k = snprintf(b, sizeof(b), "%"DIMID_FMT, res.cnt);
        res.cwd_cls = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
        if (k > n) twr_pad(twrite, (size_t)(k-n));
      }                         /* compute width of prediction column */
      if (res.col_msd) {        /* if to write a membership column */
        twr_fldsep(twrite);     /* write a field separator and */
        twr_puts(twrite, res.col_msd);      /* the column name */
        if ((mode & AS_ALIGN) && (mode & AS_WEIGHT)) {
          n = (int)strlen(res.col_msd);
          k = res.dig_msd +3;   /* compute width of membership column */
          res.cwd_msd = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
          if (k > n) twr_pad(twrite, (size_t)(k-n));
        }                       /* pad with blanks if requested */
      } }
    else {                      /* if to write a normal record */
      n = twr_printf(twrite, "%"DIMID_FMT, res.cls);
      if (res.cwd_cls > n) twr_pad(twrite, (size_t)(res.cwd_cls-n));
      if (res.col_msd) {        /* if to write a membership column */
        twr_fldsep(twrite);     /* write separator and membership */
        n = twr_printf(twrite, "%.*g", res.dig_msd,
                       cls_msdeg(clset, res.cls));
        if (res.cwd_msd > n) twr_pad(twrite, (size_t)(res.cwd_msd-n));
      }                         /* if to align, pad with blanks */
    } }
  else {                        /* if to write membership degrees */
    if (mode & AS_ATT) {        /* if to write the header */
      if (!res.col_msd) res.col_msd = "";
      m = (int)strlen(res.col_msd);
      if (mode & AS_ALIGN) {    /* if to align the column */
        n = snprintf(b, sizeof(b), "%"DIMID_FMT, res.cnt) +m;
        k = res.dig_msd +3;     /* compute width of activ. columns */
        res.cwd_msd = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
      }
      for (c = 0; c < res.cnt; c++) {
        if (c > 0) twr_fldsep(twrite);
        twr_puts(twrite, res.col_msd);
        n = m +twr_printf(twrite, "%"DIMID_FMT, c+1);
        if (res.cwd_msd > n) twr_pad(twrite, (size_t)(res.cwd_msd-n));
      } }                       /* print a column for each cluster */
    else {                      /* if to write a normal record */
      for (c = 0; c < res.cnt; c++) {  /* traverse the clusters */
        if (c > 0) twr_fldsep(twrite); /* write separator */
        n = twr_printf(twrite, "%.*g", res.dig_msd, cls_msdeg(clset,c));
        if (res.cwd_msd > n) twr_pad(twrite, (size_t)(res.cwd_msd-n));
      }                         /* for each cluster print */
    }                           /* the membership degree and */
  }                             /* pad the column for alignment */
}  /* infout() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_cls  = NULL;      /* name of cluster set file */
  CCHAR   *fn_inp  = NULL;      /* name of input  table file */
  CCHAR   *fn_out  = NULL;      /* name of output table file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  int     matinp   = 0;         /* flag for numerical matrix input */
  int     mode     = AS_ATT|AS_MARKED; /* table file read  mode */
  int     mout     = AS_ATT;           /* table file write mode */
  double  *vec;                 /* to traverse data points/vectors */
  DIMID   clscnt;               /* number of clusters */
  ATTID   a, c;                 /* number of attributes */
  DIMID   m, o;                 /* number of inputs/dimensions */
  DIMID   n, r;                 /* number of data points/vectors */
  double  w;                    /* weight of data points/vectors */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] clsfile "
                    "[-d|-h hdrfile] tabfile outfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-c#      name of the cluster index field        "
                    "(default: membership degrees)\n");
    printf("-m#      name of the membership field(s)        "
                    "(default: \"%s\")\n", res.col_msd);
    printf("-o#      digits of membership degrees           "
                    "(default: %d)\n",     res.dig_msd);
    printf("-a       align fields in output file            "
                    "(default: single separator)\n");
    printf("-w       do not write field names to output file\n");
    printf("-r#      record  separators                     "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                     "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                     "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("clsfile  file to read cluster set description from\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write extended tuples to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'c': optarg = &res.col_cls;               break;
          case 'm': optarg = &res.col_msd;               break;
          case 'o': res.dig_msd = (int)strtol(s, &s, 0); break;
          case 'a': mout  |=  AS_ALIGN;                  break;
          case 'w': mout  &= ~AS_ATT;                    break;
          case 'r': optarg = &recseps;                   break;
          case 'f': optarg = &fldseps;                   break;
          case 'b': optarg = &blanks;                    break;
          case 'C': optarg = &comment;                   break;
          case 'n': mode  |=  AS_WEIGHT;                 break;
          case 'd': mode  |=  AS_DFLT;                   break;
          case 'h': optarg = &fn_hdr;                    break;
          default : error(E_OPTION, *--s);               break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_cls = s;      break;
        case  1: fn_inp = s;      break;
        case  2: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument and */
  if (k != 3) error(E_ARGCNT);  /* the number of arguments */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  (!fn_cls || !*fn_cls) i++;
  if  (!fn_inp || !*fn_inp) i++;
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  mout |= AS_MARKED|AS_INFO1|AS_RDORD;
  if ((mout & AS_ATT) && (mout & AS_ALIGN))
    mout |= AS_ALNHDR;          /* set align to header flag */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse cluster set --- */
  scan = scn_create();          /* create a scanner */
  if (!scan) error(E_NOMEM);    /* for the cluster set */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_cls) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  matinp = (scn_first(scan) == T_ID)
        && (strcmp(scn_value(scan), "dom") != 0);
  if (matinp) {                 /* if matrix version */
    clset = cls_parse(scan,0,1);/* parse cluster descriptions */
    if (!clset || !scn_eof(scan, 1))
      error(E_PARSE, scn_name(scan));
    m = cls_inpcnt(clset);      /* get the number of inputs */
    fprintf(stderr, "[%"DIMID_FMT" input(s),", m); }
  else {                        /* if table version */
    attset = as_create("domains", att_delete);
    if (!attset) error(E_NOMEM);      /* create an attribute set */
    if ((as_parse(attset, scan, AT_ALL, 1) != 0))
      error(E_PARSE, scn_name(scan)); /* parse domain descriptions */
    a = as_attcnt(attset);      /* parse domain descriptions */
    if (a < 1) error(E_ATTCNT); /* check for at least one attribute */
    attmap = am_create(attset, 0, 1.0);
    if (!attmap) error(E_NOMEM);/* create an attribute map */
    c = as_target(attset, NULL, -1);
    if (c < a)                  /* if there is a target attribute, */
      am_target(attmap, c);     /* set it in the attribute map */
    m = am_inpcnt(attmap);      /* get the number of inputs */
    clset = cls_parse(scan,m,1);/* parse cluster descriptions */
    if (!clset || !scn_eof(scan, 1))
      error(E_PARSE, scn_name(scan));
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
  }                             /* print a log message */
  if (m < 1) error(E_INPCNT);   /* check for at least one input */
  clscnt = cls_clscnt(clset);   /* get the number of clusters */
  fprintf(stderr, " %"DIMID_FMT" cluster(s)]",   clscnt);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  cls_setup(clset);             /* set up the cluster set */
  res.cnt = clscnt;             /* note the number of clusters */

  if (matinp) {                 /* if matrix version */
    /* --- process data points/vectors --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_inp) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    twrite = twr_create();      /* create a table writer and */
    if (!twrite) error(E_NOMEM);/* open the output file */
    twr_xchars(twrite, recseps, fldseps, blanks, "");
    if (twr_open(twrite, NULL, fn_out) != 0)
      error(E_FOPEN, twr_name(twrite));
    o = -1;                     /* read the first data point */
    k = vec_readx(&vec, &o, tread);   /* from the input file */
    if (k) error(k, TRD_INFO(tread));
    if (o < m) { free(vec); error(E_VECSIZE, o); }
    for (n = 0; k == 0; n++) {  /* data point/vector read loop */
      for (c = 0; c < o; c++) { /* traverse the read vector */
        twr_printf(twrite, "%.*g", res.dig_msd, vec[c]);
        twr_fldsep(twrite);     /* print the vector elements */
      }                         /* followed by a field separator */
      c = cls_exec(clset, vec, NULL); /* execute the cluster set */
      if (res.col_cls) {        /* if to print a cluster index */
        twr_printf(twrite, "%"DIMID_FMT, c);
        if (res.col_msd) {      /* if to print membership degree */
          twr_fldsep(twrite);   /* print a separator */
          twr_printf(twrite, "%.*g", res.dig_msd, cls_msdeg(clset, c));
        } }                     /* print the membership degree */
      else {                    /* if to print all m.s. degrees */
        for (c = 0; c < clscnt; c++) {
          if (c > 0) twr_fldsep(twrite);
          twr_printf(twrite, "%.*g", res.dig_msd, cls_msdeg(clset, c));
        }                       /* print the membership degrees */
      }                         /* to the different clusters */
      twr_recsep(twrite);       /* and terminate the record */
      k = vec_read(vec, o, tread);
    }                           /* read the next data point/vector */
    free(vec);                  /* delete the vector buffer */
    if (k < 0) error(k, TRD_INFO(tread));
    trd_delete(tread, 1);       /* close the input file and */
    tread = NULL;               /* delete the table reader */
    if (twr_close(twrite) != 0) error(E_FWRITE, twr_name(twrite));
    twr_delete(twrite, 1);      /* close the output file and */
    twrite = NULL;              /* delete the table writer */
    fprintf(stderr, "[%"DIMID_FMT" vector(s)]", n);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t)); }

  else {                        /* if table version */
    /* --- read table header --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    if (fn_hdr) {               /* if a header file is given */
      t = clock();              /* start timer, open input file */
      if (trd_open(tread, NULL, fn_hdr) != 0)
        error(E_FOPEN, trd_name(tread));
      fprintf(stderr, "reading %s ... ", trd_name(tread));
      k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_close(tread);         /* read table header, close file */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", a);
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
      mode &= ~(AS_ATT|AS_DFLT);/* print a success message and */
    }                           /* remove the attribute flag */

    /* --- process table body --- */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_inp) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    if (mout & AS_ALIGN) {      /* if to align the output columns */
      table = tab_create("table", attset, tpl_delete);
      if (!table) error(E_NOMEM);  /* read the data table */
      k = tab_read(table, tread, mode);
      if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
      trd_delete(tread, 1);     /* delete the table reader */
      tread = NULL;             /* and clear the variable */
      a = tab_attcnt(table);    /* get the number of attributes */
      n = tab_tplcnt(table);    /* get the number of tuples */
      w = tab_tplwgt(table);    /* and print a success message */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
      fprintf(stderr, " %"TPLID_FMT, n);
      if (w != (double)n) fprintf(stderr, "/%g", w);
      fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
      twrite = twr_create();       /* create a table writer and */
      if (!twrite) error(E_NOMEM); /* configure the characters */
      twr_xchars(twrite, recseps, fldseps, blanks, "");
      t = clock();              /* start timer, open output file */
      if (twr_open(twrite, NULL, fn_out) != 0)
        error(E_FOPEN, twr_name(twrite));
      fprintf(stderr, "writing %s ... ", twr_name(twrite));
      if ((mout & AS_ATT)       /* write a table header */
      &&  (as_write(attset, twrite, mout, infout) != 0))
        error(E_FWRITE, twr_name(twrite));
      mout = AS_INST | (mout & ~AS_ATT);
      m += (res.col_cls ? (res.col_msd ? 2 : 1) : res.cnt);
      vec = (double*)malloc((size_t)m *sizeof(double));
      if (!vec) error(E_NOMEM); /* create a data point/vector buffer */
      for (r = 0; r < n; r++) { /* traverse the tuples */
        tpl_toas(tab_tpl(table, r));     /* store in attribute set */
        am_exec(attmap, AM_INPUTS, vec); /* and convert to vector */
        res.cls = cls_exec(clset, vec, NULL);
        if (as_write(attset, twrite, mout, infout) != 0) {
          free(vec); error(E_FWRITE, twr_name(twrite)); }
      }                         /* write the current tuple */
      free(vec); }              /* delete the vector buffer */
    else {                      /* if to process tuples directly */
      k = as_read(attset, tread, mode);     /* read table header */
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      twrite = twr_create();       /* create a table writer and */
      if (!twrite) error(E_NOMEM); /* configure the characters */
      twr_xchars(twrite, recseps, fldseps, blanks, "");
      t = clock();              /* start timer, open output file */
      if (twr_open(twrite, NULL, fn_out) != 0)
        error(E_FOPEN, twr_name(twrite));
      if ((mout & AS_ATT)       /* write a table header */
      &&  (as_write(attset, twrite, mout, infout) != 0))
        error(E_FWRITE, twr_name(twrite));
      vec = (double*)malloc((size_t)m *sizeof(double));
      if (!vec) error(E_NOMEM); /* create a data point/vector buffer */
      mout = AS_INST | (mout & ~AS_ATT);
      i = mode; mode = (mode & ~(AS_DFLT|AS_ATT)) | AS_INST;
      if (!(i & AS_DFLT))       /* if not done yet, read first tuple */
        k = as_read(attset, tread, mode);
      for (w = 0, n = 0; k == 0; n++) {
        w += as_getwgt(attset); /* sum the tuple weight */
        am_exec(attmap, AM_INPUTS, vec);    /* convert to vector and */
        res.cls = cls_exec(clset, vec, NULL); /* execute cluster set */
        if (as_write(attset, twrite, mout, infout) != 0) {
          free(vec); error(E_FWRITE, twr_name(twrite)); }
        k = as_read(attset, tread, mode);
      }                         /* try to read the next tuple */
      free(vec);                /* delete the vector buffer */
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_delete(tread, 1);     /* delete the table reader */
      tread = NULL;             /* and clear the variable */
    }
    if (twr_close(twrite) != 0) error(E_FWRITE, twr_name(twrite));
    twr_delete(twrite, 1);      /* close the output file and */
    twrite = NULL;              /* delete the table writer */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
    fprintf(stderr, " %"TPLID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
  }                             /* if (matinp) .. else .. */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
